:- module(modules, [start_module/1, end_module/1, add_exports/1,
	            load_module/1, unload_module/1, current_mod_name/1] ).

%%
:- use_module(library(lists),      [append/3] ).
:- use_module(library(dynamic),    [assert/1, retract/1] ).
:- use_module(library(aggregates), [findall/3] ).
:- use_module(library(compiler),   [unload/1]).

%% 
:- use_module(symtab, [get_rules/2, get_object_property/3,
	                            get_constructors/2,
	                            set_object_property/3,
	                            load_interfaces/1, is_type/1,
	                            unload_interfaces/1] ).
:- use_module(types,             [type_name/2]).
:- use_module(scope,             [open_scope/1] ).
:- use_module(cmdopts,           [cmd_options/2]).

%% This module implements some predicates for the module system.

:- dynamic current_module/2.

%% A module is represented like:
%%
%% curry_mod(Name, Symbols)

%% 

current_mod_name(Name) :-
	current_module(Name, _).

%% Predicates for parser.

start_module(Module) :-
	assert(current_module(Module, [])).

%% Resolve true exports and record them in the symbol table.
end_module(Module) :-
	retract(current_module(Module, Exports)),
	(
	    Exports = all ->
	    findall(id(F), (
			       get_rules(F, Rules), 
			       \+ Rules = [builtin(imported)]
			   ), NExports),
	    findall(type(Type, Cons), (
					  is_type(Type),
					  type_name(Type, tid(Module, _)),
					  (
					      get_constructors(Type, Cons) ->
					      true
					  ;
					      Cons = []
					  )
				      ), TExports),
	    append(TExports, NExports, FExports),
	    set_object_property(Module, exports, FExports)
	;
	    set_object_property(Module, exports, Exports)
	).

%% Warning, the condition inside the findall should be handled with care.
load_exports(Module, Imports) :-
	use_module(sloth(Module)),
	!,
	open_scope(imported(Module)),
	findall(iface(Iface), 
	        ( 
		    Module:exports(Iface),
		    (
			Iface = id(qid(_, Name), _, _),
			(
			    member(id(Name), Imports),
			    !			    
			;
			    Imports = []
			)
		    ;
			Iface = type(Name, _)
		    )
		), Exports),
	load_interfaces(Exports).

%% Search order.
:- multifile file_search_path/2.
:- dynamic file_search_path/2.

file_search_path(sloth, LibDir) :-
	cmd_options('-l', [LibDir]).
file_search_path(sloth, '.').

%% This is called when a import is found.
load_module(module(Name, Imports, noimport)) :-
	load_exports(Name, Imports),
	!.

load_module(module(Name, Imports, engine)) :-
	set_object_property(dummy, imports, module(Name, Imports, engine)),
	!.
	
%% Fixme, the set_object property doesn't really belong to load_module.
load_module(module(Name, Imports, _Modifiers)) :-
	load_exports(Name, Imports),
	set_object_property(dummy, imports, module(Name, Imports, _Modifiers)),
	!.

%%
add_exports(all) :-
	retract(current_module(M, _)),
	assert(current_module(M, all)).

add_exports(List) :-
	list(List),
	retract(current_module(M, L)),
	append(List, L, NewL),
	assert(current_module(M, NewL)).

unload_module(Module) :-
	unload(Module), % This kills tohnf, etc...
	unload_interfaces(Module),
	!.

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:
