:- module(typer, [check_infer_types/0, tipar/2, start/0, infer_type/2]).

:- use_module(library(lists), [intersection/3, list_concat/2]).

:- use_module(symtab, [get_rules/2, set_object_property/3,
                       get_object_property/3, get_real_rules/2,
                       replace_object_property/3]).  

:- use_module(curryast, [split_rule/3, get_params/2, get_function/2,
	                 get_identifiers/2]).
                       % solo para pruebas

:- use_module(parser, [parse/1]).


%% Operators includes.
:- include(type_ops).
:- include(ast_ops).

% librerias que se usan para este módulo
:- use_module(library(pretty_print), [pretty_print/2]).
:- use_module(library(write),        [write/1] ).

:- use_module(library(lists),        [nonsingle/1, append/3, delete/3, nocontainsx/2, difference/3]).
:- use_module(library(terms_vars)).

%% Pruebas
:- use_module(parser,     [parse_expr/2]).
:- use_module(trans_lazy, [tr_expr/2] ).
:- use_module(curryast,   [map_tree/3] ).
:- use_module(unref,      [free_vars/3] ).
:- use_module(localdefs,  [proccess_ldef_global/2, proccess_ldef_pattern/2]).
:- use_module(modules,    [load_module/1]).
:- use_module(unref,      [free_all_type_vars/0]).

%% Typer description

%% The typer operates directly with the ast representation given by the parser.
%% 
%% 1.- Generate connex components for functions that have to be typed together.
%% 2.- Sort connex parts by call order, this is done by simple identifier inclusion.
%% 3.- Type each connex part. Each one ...

%% Old description

% 6. Para tipar una componente conexa, primero se obtiene una expresion que pueda ser tipable, a
% 	partir de las reglas de las funciones que intervienen. (ver Reglas de Inferencia del predicado tipo).
% 7. Una vez obtenidas las expresiones, se utiliza el predicado tipo para obener el tipo. 
% 8. Una vez obtenido el tipo, se almacena en la TS con set_object_property(funcion,type,tipo).
% 9. El manejador de errores guardan el mensaje de error en la tabla como si fuese un tipo, y ademas
% 	escribe un mensaje de error.
% 10. Antes de almacenar el tipo en la TS, el tipador comprueba si tenia un tipo definido. De ser asi
% 	hace las comprobaciones oportunas para ver que tipo tiene que guardar, si el inferido o el definido.


%% Apolo comments about type classes.

%% 1. Si una funcion llama a un operando que pertenece a alguna clase, esta se propaga
%% 	a las variables que lo utilicen.
%% 2. Esto se consigue utilizando etiquetas.
%% 3. Si la variable con etiqueta se unifica con un Atomo, se elimina la etiqueta y se unifica.
%% 4. Si la variable con etiqueta se unifica con otra variable, si esta no tiene etiqueta, se
%% unifica con etiqueta, y si la tiene, se comprueba si la clase de la etiqueta es compatible.
%% 5. las operaciones con clases seran de la forma (== :: clase(eq,X) => X -> X -> Bool)
%%
%% Cuestiones por resolver
%% 7. si el tipo definido no tiene eq(X)=> y el inferido si, ¿debe dar error? Yo creo que no.


%% Expressions typer.

%% infer_type(Expr, Type) succeeds if Expr has type Type.
%% We need to add the type unification framework.

%% Small hack, constants already carry their type.
infer_type(const(Type, _Value), Type) :- 
	!.

%% Variables have a to be known type.
infer_type(freevar(X), X) :- !.
infer_type(param(X), X) :- !.

%% Identifiers should have a type.
infer_type(qid(M, N), Type) :-
	!,
	get_object_property(qid(M,N), type, Type).

%% For function applications.
infer_type(F@Args, Type) :-
	infer_type(F, FType),
	infer_check_fapp(FType, Args, Type).

%% infer_check_fapp checks a function application.
infer_check_fapp(FType, [], FType) :- !.
infer_check_fapp(FType, [A|AS], RType) :-
	FType = (T1 --> T2),
	infer_type(A, T1),
	!,
	infer_check_fapp(T2, AS, RType).

%% I think all these are obsolete.

%% This is for lists. Ugly hack, has to die.
% tipo([X|[]], list(Y)) :-
%  	(
%  	    tipo(X, Y) 
%  	; 
%  	    throw(error_typer('Different types in list'))
%  	). 

% tipo([X|Xs], list(Y)) :-
% 	tiplista(X, Y),
%  	Xs\=[],
%  	tipo(Xs, list(Y)).


% tipo tupla
% tipo((X, Y), (A, B)) :-
% 	tipo(X, A), 
% 	tipo(Y, B).

% % abstracciones
% tipo(lambda(P, E), (T1-->T2)) :-
% 	mis1(P, T1), 
% 	mis1(E, T2).

% tipo(case([]), _X).
% tipo(case([L1|Ls]), (T1-->T2)) :-
% 	mis2(L1, (T1-->T2)), % o debe ser mis1 ??
% 	tipo(case(Ls), (T1-->T2)).

% % Let (¿Esto ya no vale?)
% tipo(le(X, E1, E2), T) :-
% 	tipo(E1, T1), 
% 	tipo(X, T1), 
% 	tipo(E2, T).

% % Fixed Point
% tipo(fix(X, E), T) :-
% 	tipo(X, T), 
% 	tipo(E, T).

% tiplista(X, Y) :-
% 	(
% 	    tipo(X, Y) 
% 	; 
% 	    throw(error_typer('Different types in list'))
% 	). 

% % para comprobar errores type Mismatch
% mis1(P, T1) :-
% 	tipo(P, T1) 
%  ; 
% 	throw(error_typer('Type Mismatch')).


% mis2(P, T1) :-
% 	tipo(P, T1) 
%  ; 
% 	throw(error_typer('Rules have diferrent types')).

%% Module typer.

%% The approach used for typing is to build a dependency tree of all
%% functions and start typing the leaves.

%% This is not going to be successful in the case of recursive
%% functions, so we use fixed point typing.

%% The steps are:

%% 1.- Generate a connex components for all the rules, I mean, the set
%%     of set of functions that are called between.

%% 2.- Order each set based in the call order. In case there's no
%%     order we generate a fixed point equation.

%% 3.- Type each set.

%% R is the list of lists of function names that are in the same
%% connex group.

gen_connex(R) :- % 
	findall(N, get_real_rules(N, _), L),
	build_connex(L, [], R).

build_connex([], L, L).
build_connex([F|FS], P, N) :-
	get_functions(F, Ids),
	add_to_existing(F, Ids, P, PM),
	!,
	build_connex(FS, PM, N).
	
build_connex([F|FS], P, N) :-
	!,
	build_connex(FS, [[F]|P], N).

%% Warning, the sets will have repeated elements if the function is recursive.
%% Warning, this should only care of the module's functions.
%% No set, we have to create the first
add_to_existing(Fun, _FunIds, [], [[Fun]]) :-
	!.

%% The function is already in a set.
add_to_existing(Fun, _FunIds, [CGroup|Rest], [CGroup|Rest]) :-
	member(Fun, CGroup),
	!.

%% The function belongs to that set, but it's not there.
add_to_existing(Fun, FunIds, [CGroup|Rest], [NewGroup|Rest]) :-
	intersection(FunIds, CGroup, [_|_]),
	!,
	NewGroup = [Fun|CGroup].

%% 
add_to_existing(Fun, FunIds, [CGroup|Rest], [CGroup|DRest]) :-
	add_to_existing(Fun, FunIds, Rest, DRest).
	
%% 
get_functions(F, Ids) :-
	get_rules(F, Rules),
	all_rules_fun(Rules, RIds),
	list_concat(RIds, Ids).

all_rules_fun([], []).
all_rules_fun([R|RS], [Ids|IS]) :-
	get_identifiers(R, AllIds),
	only_functions(AllIds, Ids),
	all_rules_fun(RS, IS).

only_functions([], []).
only_functions([F|FS], RS) :-
	get_object_property(F, constructs, _),
	!,
	only_functions(FS, RS).

only_functions([F|FS], [F|RS]) :-
	!,
	only_functions(FS, RS).

%% Check all types in the symbol table.
check_infer_types :-
	ordeneval(L), %% Obtiene las reglas de evaluación.
	tipear(L,_L1).

%% Function typing.
%% To type a function we have to type each rule.


% Generates ordered list of conex components ready for typing.
ordeneval(L) :- 
	ordenar(_R, [], L).  

tipear([],[]).

% para calcular funcions que llaman a otras funciones
% apunta(N,E) <-- N apunta a E.
apunta(N, E) :- 
	regla(N, _, Y), 
	N\=E,  
	esta(E, Y).

% apunta, si el segundo término es una lista.
% para ver si una componente conexa apunta a una función).
apunta2(N, [A|[]]) :- 
	apunta(N, A).
apunta2(N, [X|Xs]) :- 
	apunta(N, X), 
	Xs\=[], ! 
 ;      apunta2(N, Xs).

% apunta, si el primer término es una lista.
% (para ver si una funcion apunta a una componente conexa).
apunta3([X|[]], Y):- 
	apunta(X, Y).
apunta3([X|Xs], Y) :- 
	apunta(X, Y), 
	Xs\=[], ! 
 ;      apunta3(Xs, Y).

% apunta si los dos términos son listas (para ver si dos componentes
% conexas se apuntan)
apunta1([X|[]], [Y|Ys]) :- 
	apunta2(X, [Y|Ys]).
apunta1([X|Xs], [Y|Ys]) :- 
	apunta2(X, [Y|Ys]), 
	Xs\=[], ! 
 ;      apunta1(Xs, [Y|Ys]).

%        convierte las reglas en una sola expresión con lambdas.
%        ------------------------------------------------------

elambdaux(_N, S, [], S).
elambdaux(N, lambda(X, E), [X|Xs], S) :-
	elambdaux(N, E, Xs, S).

elambda(N, E) :- 
	regla(N, L, S), elambdaux(N, E, L, S).
 

% modificamos ecase para que sólo ponga case si hay varias reglas
ecase(N, case(L)) :-
	setof(X, elambda(N, X), L), 
        nonsingle(L).     %	\+ hayuna(L).
ecase(N, X) :-
	setof(X, elambda(N, X), L), 
	\+ nonsingle(L),       % hayuna(L), 
	L=[X|[]].

lfix([],[]).
lfix([X|Xs], [Y|R]) :- 
	ecase(X, Y), 
	lfix(Xs, R).


% saca la expresión con fix de una componente conexa y sustituye los
% nombre de reglas por var(X) en los fix, y lo convierte en tuplas.

%% egallego: Esta es dura. Creo que lo que hay que hacer aquí es
%% sustituir el nombre de las funciones dentro de las componentes
%% conexas por variables, y tiparlas de esa manera.

%% Pregunta: ¿Cuando determinamos los tipos de una función?

%% XXX Documentar el formato que usaba.

ffix2(L, L1) :-
	\+ nonsingle(L),  % hayuna(L), 
	L=[A], 
	\+esrecur(A), 
	ecase(A, L1), 
	!.
ffix2(L, R) :-
	\+ nonsingle(L),  % hayuna(L),
	L=[A], 
	esrecur(A), 
	ecase(A,B), 
	sust(A, fix(A, B), R), 
	!.
ffix2(L, fix(X, Y)) :-
	nonsingle(L), % \+hayuna(L), 
	lfix(L,L1),
        reemplazo(L, fix(L, L1), R), 
	R=fix(A,B), 
	listatupla(A, X), 
	listatupla(B, Y),
	!.



%  predicados para convertir var(x) en var(X) en las reglas         
% ---------------------------------------------------------
  
% unir une dos listas sin elementos repetidos

unir([], [], []) :- !.
unir([], Y, Y) :- !.
unir(X, [], X) :- !.
unir([X|Xs], Y, [X|W]) :- 
	nocontainsx(Y, X),    %noesta(X, Y), 
	unir(Xs, Y, W).
unir([X|Xs], Y, W) :- 
        \+ nocontainsx(Y,X),   %	\+ noesta(X, Y), 
	unir(Xs, Y, W).
   
%               Calculo de componentes conexas para fix
%               ---------------------------------------

% predicado para ver si un nombre aparece en una expresion
%% XXX Lo tenemos hecho ya.
esta(N, N).
esta(N, lambda(X, Y)) :- 
	esta(N, X), ! 
 ;      esta(N, Y), !.
esta(N, (X@Y)) :- 
	esta(N, X), ! 
 ;      esta(N, Y), !.
esta(N, le(X, Y, Z)) :- 
	esta(N, X), !
 ;      esta(N, Y), ! 
 ;      esta(N, Z), !.
esta(N, (X, Y)) :- 
	esta(N, X), ! 
 ;      esta(N, Y), !.

% predicado para sustituir el nombre de la regla por una var(X)
% y usarlo en el tipado de fix (punto fijo)
%% Esta hay que mirarla.
sust(N, N, var(_X)).
sust(N, qid(X, Y), qid(X, Y)) :-
	N\=qid(X,Y).
sust(N, lambda(A, B), lambda(J, K)) :- 
	sust(N, A, J), 
	sust(N, B, K).
sust(N, (A@B), (J@K)) :- 
	sust(N, A, J),
	sust(N, B, K).

/*
sust(N, le(A, B, C), le(J, K, L)) :- 
	sust(N, A, J),
	sust(N, B, K),
	sust(N, C, L).

*/
sust(N, (A, B), (J, K)) :- 
	sust(N, A, J),
	sust(N, B, K).
sust(N, B, B) :- 
	atom(B), 
	\+ list(B), 
	N\=B.
sust(_N, B, B) :- 
	number(B).
sust(_N, var(A), var(A)).
sust(_N, const(A,B), const(A,B)).        % EMILIO

% estas siguen haciendo falta?
sust(_N, num(X), num(X)).
sust(_N, id(X), id(X)).
sust(_N, char(X), char(X)).
sust(_N, str(X), str(X)).

sust(N, L, L1) :- 
	list(L), 
	sustl(N, L, L1).
sust(N, case(X), case(Y)) :- 
	sust(N, X, Y).
sust(N, fix(X, Y), fix(A, B)) :- 
	sust(N, X, A),
	sust(N, Y, B).

% como sust, pero para listas
sustl(_N, [], []).
sustl(N, [X|Xs], [A|As]) :- 
	sust(N, X, A), 
	sustl(N, Xs, As).
 
% reemplaza todas las variables en un fix([],[]) (A debe ser fix)    
reemplazo([], A, A).
reemplazo([X|Xs], A, Z) :- 
	sust(X, A, B), 
	reemplazo(Xs, B, Z).

reemplazo2(fix(A, B), C) :- 
	reemplazo(A, fix(A, B), C).
   
% predicado para obtener la expresión final tipable
efinal(N, fix(N, case(L))) :- 
	elambda(N, Y), 
	esta(N, Y),
        setof(X, elambda(N, X), L), !.
efinal(N, case(L)) :- 
	setof(X, elambda(N, X), L).

% para ver si una función se llama a si misma
esrecur(X) :- 
	regla(X, _, Y), 
	esta(X, Y).




%                    Predicados para Listas
% algunas se podrían sustituir por las que están en la libreria list
% (programadas en un momento en el que no se disponia de la documentación
% de la libreria list)
%  -----------------------------------------------

  
% convierte una lista en una tupla
listatupla([A|As], (A, Bs)) :- 
	listatupla(As, Bs).
listatupla([X|[]], X).

/*

% indica si un elemento NO ESTA en una lista. (es como nocontainsx(L,X) ).
noesta(A, []).
noesta(A, [X|Xs]) :- 
	A\=X, 
	noesta(A, Xs).

% elimina un elemento de una lista (es como delete(L1,E,L2) )
quitar(A, [], []).
quitar(A, [A|Xs], Ys) :- 
	quitar(A, Xs, Ys).
quitar(A, [X|Xs], [X|Ys]) :- 
	A\=X, 
	quitar(A, Xs, Ys).

% elimina los elementos de una lista de otra lista  (es como difference(Linicial,Lquitar,Lfinal)
quitarl([], Xs, Xs).
quitarl([A|As], Xs, Zs) :- 
	quitar(A, Xs, Ys), 
	quitarl(As, Ys, Zs).

*/

% ordena una lista de funciones que se apuntan unas a otras
ordenar([], Xs, Xs).
ordenar([A|As], Bs, Zs) :- 
	inserta(A, Bs, Cs), 
	ordenar(As, Cs, Zs).

% inserta una funcion dependiendo de si apunta o no a las demás
inserta(A, [], [A]).
inserta(A, [B|Bs], [A|Cs]) :- 
	apunta1(B, A), 
	Cs=[B|Bs].
inserta(A, [B|Bs], [B|Cs]) :- 
	\+ apunta1(B, A), 
	inserta(A, Bs, Cs).

% devuelve una lista sin elementos repetidos 
norep([],[]).
norep([X|[]], [X|[]]).
norep([X|Xs], [X|Zs]) :- 
	Xs\=[], 
	delete(Xs, X, Ys),    % quitar(X, Xs, Ys), 
	norep(Ys, Zs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%  TRATAMIENTO DE ERRORES

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% función para comprobar si el tipo inferido
%% concuerda con el especificado.

%% comprobar_tipos

% si el tipo inferido y el definido son iguales, todo queda igual
% si el tipo inferido es más restrictivo, da error
% si el tipo definido es el más restrictivo, se da un warning 
%  y se usa este tipo.



check_type((Regla,InfrType)) :-
	( get_object_property(Regla,type,DefType),	 
	 catch(comprobar_tipo(Regla,InfrType,DefType),error_typer(Mensaje), 
	         (write(Regla),write(' - '),write(Mensaje), write('\n')))
	)
 ;
	set_object_property(Regla,type,InfrType).

comprobar_tipo(Regla, InfrType, DefType) :- 
	( numbervars(DefType,  0, _),
	  numbervars(InfrType, 0, _),
	  InfrType==DefType
	)  
	% el tipo inferido y el definido son iguales,
        % dejamos en la tabla el tipo definido.    
 ;
	comprobar_tipo2(Regla,InfrType,DefType).
	
comprobar_tipo2(Regla,InfrType,DefType) :-
	(
	InfrType=DefType,  % unificamos los tipos
	comprobar_tipo3(Regla,DefType)
        )
 ;
	% el tipo inferido y el definido son diferentes
        %write('error1').
	throw(error_typer('Inferred type does not match the specified one')). % typerror(Regla,11)
	
comprobar_tipo3(Regla,X) :-
	( %%  lookup(Regla,defType,DefType),
	get_object_property(Regla,type,DefType),
	numbervars(DefType,0,_),
	numbervars(X,0,_),
	DefType==X,
	% El tipo inferido es más restrictivo, lo sustituimos
	% y enviamos un warning
	replace_object_property(Regla,type,X),
        %write('error warning')
        
	throw(error_typer('WARNING: Inferred type does not match the specified one'))
	)
 ;
        write('error3').
	% throw(error_typer('Inferred type does not match the specified one')) %  typerror(Regla,11) 
	
                   
/*

% Lista de errores de tipos
mensaje_error(_,Num_Error,Mensaje):-
	Num_Error=1, Mensaje='Undefined Symbol'  % no se usa
 ;          Num_Error=2, Mensaje='Different types in list'
 ;          Num_Error=3, Mensaje='Type Mismatch'
 ;          Num_Error=4, Mensaje='Undefined Constructor'  % No se usa
 ;          Num_Error=5, Mensaje='A constraint type was expected'
 ;          Num_Error=6, Mensaje='Then and Else expressions have different types' % seria type mismatch
 ;          Num_Error=7, Mensaje='If expression must be boolean'  % sería type mismatch
 ;          Num_Error=8, Mensaje='Invalid constraint in choice expression'
 ;          Num_Error=9, Mensaje='Expression in choice are not the same type'
 ;          Num_Error=10, Mensaje='Rules have different types'
 ;          Num_Error=11, Mensaje='Inferred type does not match the specified one'
 ;          Num_Error=12, Mensaje='Infinite type for function'
 
% A partir de aqui van los warnings

    ;          Num_Error=13, Mensaje='Inferred type does not match the specified one'
    ;          Num_Error>13, Mensaje=0
    ;          Num_Error<1,  Mensaje=0.


*/

start :-
	load_module(module('Prelude', [], noimport)).

tipar(Input, Tipo) :-
	catch(parse_expr(Input, Expr), E, write(E)),
	map_tree(Expr, proccess_ldef_global, Expr1),
	map_tree(Expr1, proccess_ldef_pattern, LExpr),
	map_tree(LExpr, free_vars(_Dicc), ExprFree),
	infer_type(ExprFree, Tipo).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

